/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                           MSX.c                         **/
/**                                                         **/
/** This file contains implementation for the MSX-specific  **/
/** hardware: slots, memory mapper, PPIs, VDP, PSG, clock,  **/
/** etc. Initialization code and definitions needed for the **/
/** machine-dependent drivers are also here.                **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994,1995                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

#include "MSX.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define PRINTOK     if(Verbose) printf(MsgOK)
#define PRINTFAILED if(Verbose) printf(MsgFAILED)

char *MsgOK = "OK\n";
char *MsgFAILED = "FAILED\n";

byte Verbose     = 1;               /* Debug msgs ON/OFF      */
byte SaveCMOS    = 0;               /* Save CMOS.ROM on exit  */
byte MSXVersion  = 0;               /* 0=MSX1, 1=MSX2         */
byte RAMPages=4,VRAMPages=4;        /* Number of RAM pages    */

byte UPeriod = 1;        /* Number interrupts / screen update */

char *CartA = "CARTA.ROM";          /* Cartridge A ROM file   */
char *CartB = "CARTB.ROM";          /* Cartridge B ROM file   */

char *PrnName = NULL;               /* Printer redirect. file */
FILE *PrnStream = NULL;

byte *VRAM=NULL,*RAM=NULL,*VPAGE;   /* Main and Video RAMs    */

byte *MemMap[4][4][4];   /* Memory maps [PPage][SPage][Addr]  */

byte *RAMMap[16];                   /* Blocks for RAM Mapper  */
byte RAMMapper[4];                  /* RAM Mapper state       */
byte *ROMMap[2][256];               /* Blocks for ROM Mappers */
byte ROMMapper[2][4];               /* ROM Mappers state      */ 

byte EnWrite[4] = {0,0,0,0};        /* 1 if write enabled     */

byte PSL[4],SSL[4];                 /* Lists of current slots */
byte PSLReg,SSLReg;      /* Storage for A8h port and (FFFFh)  */

byte KeyMap[16] =                   /* Keyboard map           */
{
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF
};

byte IOReg;                         /* Storage for AAh port   */
byte PPIReg;                        /* PPI control register   */

byte PSGReg=0;                      /* PSG register number    */ 
byte PSG[16];                       /* PSG registers          */
int  Volume[3],Freq[3];             /* Sound parameters [3ch] */

byte RTCReg=0,RTCMode=0;            /* RTC register numbers   */
byte RTC[4][13];                    /* RTC registers          */

byte *ChrGen,*ChrTab,*ColTab;       /* VDP tables [screens]   */
byte *SprGen,*SprTab,*SprCol;       /* VDP tables [sprites]   */
byte Palette[32][3];                /* Palette registers      */
pair WVAddr,RVAddr;      /* Storage for VRAM addresses in VDP */
byte VKey=1,PKey=1;                 /* Status keys for VDP    */
byte FGColor=0,BGColor=0;           /* Colors                 */
byte XFGColor,XBGColor;             /* Second set of colors   */
byte ScrMode=0;                     /* Current screen mode    */
byte VDP[64],VDPStatus[16];         /* VDP registers          */


/******** Screen Mode Handlers [number of screens + 1]: *******/
struct
{                                   /* Screen mode handlers:  */
  int (*Init)();                    /* Initialize screen      */
  void (*Colors)();                 /* Update colors/palette  */
  void (*Refresh)();                /* Refresh screen         */
  byte R2,R3,R4,R5;                 /* Masks for table addr.  */
} SCR[MAXSCREEN+2] =                /* Number of screens + 1  */
{
  { InitScrF,ColorsScrF,RefreshScr0,0x7F,0x00,0x3F,0x00 }, /* SCR 0:TEXT 40x24  */
  { InitScrF,ColorsScrF,RefreshScr1,0x7F,0xFF,0x3F,0xFF }, /* SCR 1:TEXT 32x24  */
  { InitScrF,ColorsScrF,RefreshScr2,0x7F,0x80,0x3C,0xFF }, /* SCR 2:BLK 256x192 */
  { InitScrF,ColorsScrF,RefreshScr3,0x7F,0xFF,0x00,0xFF }, /* SCR 3:64x48x16    */
  { InitScrF,ColorsScrF,RefreshScr4,0x7F,0x80,0x3C,0xFF }, /* SCR 4:BLK 256x192 */
  { InitScrF,ColorsScrF,RefreshScr5,0x60,0x00,0x00,0xFC }, /* SCR 5:256x192x16  */
  { InitScrF,ColorsScrF,RefreshScr6,0x60,0x00,0x00,0xFC }, /* SCR 6:512x192x4   */
  { InitScrF,ColorsScrF,RefreshScr7,0x20,0x00,0x00,0xFC }, /* SCR 7:512x192x16  */
  { InitScrF,ColorsScrF,RefreshScr8,0x20,0x00,0x00,0xFC }, /* SCR 8:256x192x256 */
  { InitScrF,ColorsScrF,RefreshTx80,0x7C,0xF8,0x3F,0x00 }  /* SCR 0:TEXT 80x24  */
};


/*** Data for the CPU->VRAM function of V9938: ***/
struct
{
  byte *P,Mode,TR,Mask,Op,Diff,Shift[4];
  int X,NX,NY,XC,NXC,XStep,YStep;
} MMC;


/*** Load a .ROM file into given slot ***/
byte *LoadROM(char *STR,FILE *F,int P,int S,int A);
int LoadCart(char *Name,int Slot);/*** Load cartridge from a .ROM file ***/
void VDPOut(byte Reg,byte Value); /*** Write value into a VDP register ***/
void PSGOut(byte R,byte V);       /*** Write value into a PSG register ***/
void Printer(byte V);             /*** Send a character to a printer   ***/
void CheckSprites(void);      /*** Check for collisions and 5th sprite ***/
byte RTCIn(byte R);               /*** Read RTC registers              ***/
void DoVDP(byte V);               /*** Perform V9938 operation         ***/
/*** Function used by DoVDP() for operations on pixels ***/ 
byte Oper(byte A,byte B,byte N);


/****************************************************************/
/*** Write number into given IO port.                         ***/
/****************************************************************/
void DoOut(register byte Port,register byte Value)
{
  static byte Pal,Addr;     /* Palette and address buffers */
  register byte I,J,K,*P;  

  switch(Port)
  {

case 0x91: Printer(Value);break;
case 0x98: if(VKey)
           {
             if(VPAGE) *(VPAGE+WVAddr.W)=Value;
             WVAddr.W=(WVAddr.W+1)&0x3FFF;
             if(!WVAddr.W&&(ScrMode>3)) 
             {
               VDP[14]=(VDP[14]+1)&0x07;
               VPAGE=(VDP[14]<VRAMPages)?
                 VRAM+((long)VDP[14]<<14):NULL;
             }
           }
           return;
case 0x99: if(VKey) { Addr=Value;VKey--; }
           else
           {
             VKey++;
             switch(Value&0xC0)
             {
               case 0x80: VDPOut(Value&0x3F,Addr);break;
               case 0x40: WVAddr.B.l=Addr;WVAddr.B.h=Value&0x3F;break;
               case 0x00: RVAddr.B.l=Addr;RVAddr.B.h=Value;
             }
           }
           return;
case 0x9A: if(PKey) { Pal=Value;PKey--; }
           else
           {
             PKey++;J=VDP[16];
             VDP[16]=(VDP[16]+1)&0x0F;
             Palette[J][0]=(Pal&0x70)<<1;
             Palette[J][1]=(Value&0x07)<<5;
             Palette[J][2]=(Pal&0x07)<<5;
             SCR[ScrMode].Colors();
           }
           return;
case 0x9B: J=VDP[17]&0x3F;
           if(J!=17) VDPOut(J,Value);
           if(!(VDP[17]&0x80)) VDP[17]=(J+1)&0x3F;
           return;
case 0xA0: PSGReg=Value&0x0F;return;
case 0xA1: PSGOut(PSGReg,Value);return;
case 0xAA: IOReg=Value;return;
case 0xAB: if(Value&0x80) PPIReg=Value;
           else
             if(Value&0x01) IOReg|=1<<(7&(Value>>1));
             else           IOReg&=~(1<<(7&(Value>>1)));
           break;
case 0xB4: RTCReg=Value&0x0F;break;
case 0xB5: if(RTCReg<13)
           {
             J=RTCMode&0x03;RTC[J][RTCReg]=Value;
             if(J>1) SaveCMOS=1;
           }
           else if(RTCReg==13) RTCMode=Value;
           break;
case 0xA8: if(Value!=PSLReg)
           {
             for(J=0,PSLReg=Value;J<4;J++,Value>>=2)
             {
               I=Value&3;
               if(PSL[J]!=I)
               {
                 P=RAM+((word)J<<14);
                 if(EnWrite[J]) memcpy(MemMap[3][2][J],P,PAGESIZE);
                 PSL[J]=I;K=SSL[J];
                 if(MemMap[I][K][J])
                   memcpy(P,MemMap[I][K][J],PAGESIZE); 
                 else
                   memset(P,NORAM,PAGESIZE);
                 EnWrite[J]=(I==3)&&(K==2)&&MemMap[3][2][J];
               }  
             }
             if(PSL[3]==3) *(RAM+0xFFFF)=~SSLReg;
           }
           break;
case 0xFC:
case 0xFD:
case 0xFE:
case 0xFF: I=Port-0xFC;Value&=0x0F;
           if((PSL[I]==3)&&(SSL[I]==2))
           {
             P=RAM+((word)I<<14);
             if(EnWrite[I]) memcpy(MemMap[3][2][I],P,PAGESIZE);
             EnWrite[I]=RAMMap[Value]? 1:0;
             if(EnWrite[I]) memcpy(P,RAMMap[Value],PAGESIZE);
             else           memset(P,NORAM,PAGESIZE);
           }
           MemMap[3][2][I]=RAMMap[Value];RAMMapper[I]=Value;
           break; 

  }
}



/****************************************************************/
/*** Read number from given IO port.                          ***/
/****************************************************************/
byte DoIn(register byte Port)
{
  switch(Port)
  {

case 0x90: return(0xFD);                     /* Printer READY signal */
case 0x98: Port=VPAGE? *(VPAGE+RVAddr.W):255;/* VRAM read port       */
           RVAddr.W=(RVAddr.W+1)&0x3FFF;
           if(!RVAddr.W&&(ScrMode>3))
           {
             VDP[14]=(VDP[14]+1)&0x07;
             VPAGE=(VDP[14]<VRAMPages)?
               VRAM+((long)VDP[14]<<14):NULL;
           }
           return(Port);
case 0x99: Port=VDPStatus[VDP[15]];          /* VDP status registers */
           if(!VDP[15]) VDPStatus[0]&=0x5F;
           if(VDP[15]==2) VDPStatus[2]^=0x60;
           return(Port);
case 0xA2: if(PSGReg>13) return(NORAM);      /* PSG registers        */
           else return(PSG[PSGReg]);
case 0xA8: return(PSLReg);                   /* Primary slot state   */
case 0xA9: return(KeyMap[IOReg&0x0F]);       /* Keyboard port        */
case 0xAA: return(IOReg);                    /* General IO register  */
case 0xAB: return(PPIReg);                   /* PPI control register */
case 0xB5: return(RTCIn(RTCReg));            /* RTC registers        */
case 0xFC:                                   /* Mapper page at 0000h */
case 0xFD:                                   /* Mapper page at 4000h */
case 0xFE:                                   /* Mapper page at 8000h */
case 0xFF: return(RAMMapper[Port-0xFC]|0xF0);/* Mapper page at C000h */
default:   return(NORAM);                    /* If no port, ret FFh  */ 

  }
}



/****************************************************************/
/*** Switch ROM Mapper pages. This function is supposed to be ***/
/*** to be called when ROM page registers are written to.     ***/
/****************************************************************/
void MapROM(register word Addr,register byte Value)
{
  byte *P,I,J;

  J=Addr>>14;

  if((PSL[J]==1)&&ROMMap[0][0]) I=0;
  else 
    if((PSL[J]==2)&&ROMMap[1][0]) I=1;
    else
    { if(EnWrite[J]) *(RAM+Addr)=Value;return; }

  if(MemMap[I+1][0][J])
  { 
    P=MemMap[I+1][0][J];
    J=(Addr-0x4000)>>13;
    ROMMapper[I][J]=Value;

    if(ROMMap[I][Value])
    {
      memcpy(P+(Addr&0x2000),ROMMap[I][Value],PAGESIZE/2);
      memcpy(RAM+(Addr&0xE000),ROMMap[I][Value],PAGESIZE/2);
    }
    else
    {
      memset(P+(Addr&0x2000),NORAM,PAGESIZE/2);
      memset(RAM+(Addr&0xE000),NORAM,PAGESIZE/2);
    }
  }
}



/****************************************************************/
/*** Switch secondary memory slots. This function is supposed ***/
/*** to be called when value in (FFFFh) changes.              ***/
/****************************************************************/
void SSlot(register byte Value)
{
  register byte I,J,K,*P;
  
  if(PSL[3]!=3)
  { if(EnWrite[3]) *(RAM+0xFFFF)=Value;return; }

  if(SSLReg!=Value)
  {
    SSLReg=Value;

    for(J=0;J<4;J++,Value>>=2)
    {
      I=Value&3;K=SSL[J];SSL[J]=I;
      if((K!=I)&&(PSL[J]==3))
      {
        P=RAM+((word)J<<14);
        if(EnWrite[J])      memcpy(MemMap[3][K][J],P,PAGESIZE);
        if(MemMap[3][I][J]) memcpy(P,MemMap[3][I][J],PAGESIZE);
        else                memset(P,NORAM,PAGESIZE);
        EnWrite[J]=(I==2)&&MemMap[3][I][J];
      }
    }
    *(RAM+0xFFFF)=~SSLReg;
  }
}
  
  

/****************************************************************/
/*** Allocate memory, load ROM images, initialize mapper, VDP ***/
/*** CPU and start the emulation. This function returns 0 in  ***/
/*** the case of failure.                                     ***/
/****************************************************************/
int StartMSX(void)
{
  word A;
  FILE *F;
  reg R;
  int *T,I,J,K;

  /*** CMOS ROM default values:     ***/ 
  static byte RTCInit[4][13]  = 
  {
    { 0,0,0,0,0,0,0,0,0,0,0,0,0 },
    { 0,0,0,0,0,0,0,0,0,0,0,0,0 },
    { 0,0,0,0,40,80,15,4,4,0,0,0,0 },
    { 0,0,0,0,0,0,0,0,0,0,0,0,0 }
  };
  /*** PSG initial register states: ***/
  static byte PSGInit[16]  = { 0,0,0,0,0,0,0,0xFD,0,0,0,0,0,0,0xFF,0 };
  /*** VDP status register states:  ***/
  static byte VDPSInit[16] = { 0x80,0,0x4C,0,0,0,0,0,0,0,0,0,0,0,0,0 };
  /*** VDP control register states: ***/
  static byte VDPInit[64]  =
  {
    0x00,0x10,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
  };
  /*** Initial palette: ***/
  static byte PalInit[16][3] =
  {
    {0x00,0x00,0x00},{0x00,0x00,0x00},{0x20,0xC0,0x20},{0x60,0xE0,0x60},
    {0x20,0x20,0xE0},{0x40,0x60,0xE0},{0xA0,0x20,0x20},{0x40,0xC0,0xE0},
    {0xE0,0x20,0x20},{0xE0,0x60,0x60},{0xC0,0xC0,0x20},{0xC0,0xC0,0x80},
    {0x20,0x80,0x20},{0xC0,0x40,0xA0},{0xA0,0xA0,0xA0},{0xE0,0xE0,0xE0},
  };

  /*** STARTUP CODE starts here: ***/

  T=(int *)"\01\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
#ifdef LSB_FIRST
  if(*T!=1)
  {
    printf("********** This machine is high-endian. **********\n");
    printf("Take #define LSB_FIRST out and compile fMSX again.\n");
    return(0);
  }
#else
  if(*T==1)
  {
    printf("********* This machine is low-endian. **********\n");
    printf("Insert #define LSB_FIRST and compile fMSX again.\n");
    return(0);
  }
#endif

  if((RAMPages<(MSXVersion? 8:4))||(RAMPages>16))
    RAMPages=VRAMPages=MSXVersion? 8:4;
  if((VRAMPages<(MSXVersion? 4:2))||(VRAMPages>8))
    VRAMPages=MSXVersion? 4:2;

  for(I=0;I<16;I++)  RAMMap[I]=NULL;
  for(I=0;I<256;I++) ROMMap[0][I]=ROMMap[1][I]=NULL;

  for(I=0;I<4;I++)
    for(J=0;J<4;J++)
      for(K=0;K<4;K++)
        MemMap[I][J][K]=NULL;

  if(Verbose) printf("Allocating 64kB for address space...");
  if(!(RAM=malloc(4*PAGESIZE))) return(0);
  if(Verbose) printf("OK\nAllocating %dkB for VRAM...",VRAMPages*16);
  if(!(VRAM=malloc(VRAMPages*PAGESIZE))) return(0);
  if(Verbose) printf("OK\nAllocating %dx16kB RAM pages...",RAMPages);
  for(J=0;J<RAMPages;J++) 
    if(!(RAMMap[J]=malloc(PAGESIZE))) return(0);

  if(Verbose)
    printf("OK\nLoading MSX%c ROMs:\n",MSXVersion? '2':'1');
  if(MSXVersion)
  {
    if(Verbose) printf("  Opening MSX2EXT.ROM...");
    if(F=fopen("MSX2EXT.ROM","rb"))
    {
      if(LoadROM("OK\n    MSX2 ROM Extension",F,3,1,0)) J=1;
      fclose(F);if(!J) return(0);   
    }
    else { PRINTFAILED;return(0); }
    if(Verbose) printf("  Opening MSX2.ROM...");
    F=fopen("MSX2.ROM","rb");
  } 
  else
  {
    if(Verbose) printf("  Opening MSX.ROM...");
    F=fopen("MSX.ROM","rb");
  }
  if(F)  
  {
    if(Verbose) printf("OK\n");
    J=0;
    if(LoadROM("    BIOS  ROM",F,0,0,0))
      if(LoadROM("    BASIC ROM",F,0,0,1)) J=1;
    fclose(F);if(!J) return(0);J=0;
  }
  else { PRINTFAILED;return(0); }

  if(F=fopen("CMOS.ROM","rb"))
  {
    if(Verbose) printf("Found CMOS.ROM: loading...");
    if(fread(RTC,1,sizeof(RTC),F)!=sizeof(RTC))
    { PRINTFAILED;fclose(F);F=0; }
  }
  if(F) { fclose(F);PRINTOK; }
  else memcpy(RTC,RTCInit,sizeof(RTC));

  if(F=fopen("DISK.ROM","rb"))
  { LoadROM("Found DISK.ROM",F,3,1,1);fclose(F); }
  if(F=fopen("RS232.ROM","rb"))
  { LoadROM("Found RS232.ROM",F,3,3,1);fclose(F); }

  if(F=fopen("PAINTER.ROM","rb"))
  {
    char S[10];
    if(Verbose) printf("Found PAINTER.ROM:\n");
    for(J=0;J<4;J++)
    { sprintf(S,"  Part %d",J);LoadROM(S,F,1,0,J); }
    fclose(F);
  }

  LoadCart(CartA,1);LoadCart(CartB,2);

  if(!PrnName) PrnStream=stdout;
  else
  {
    if(Verbose) printf("Redirecting printer output to %s...",PrnName);
    if(!(PrnStream=fopen(PrnName,"wb"))) PrnStream=stdout;
    printf((PrnStream==stdout)? MsgFAILED:MsgOK);
  }

  if(Verbose) printf("Initializing memory mappers...");
  for(PSLReg=0x00,J=0;J<4;J++)
  { 
    PSL[J]=0;MemMap[3][2][J]=RAMMap[J];
    ROMMapper[0][J]=ROMMapper[1][J]=RAMMapper[J]=J;
  }
  for(SSLReg=0x00,J=0;J<4;J++) SSL[J]=0;  
  memcpy(RAM,MemMap[0][0][0],PAGESIZE);
  memcpy(RAM+0x4000,MemMap[0][0][1],PAGESIZE);
  memset(RAM+0x8000,NORAM,2*PAGESIZE);

  if(Verbose) printf("OK\nInitializing VDP, PSG, and CPU...");
  memcpy(VDP,VDPInit,sizeof(VDP));
  memcpy(VDPStatus,VDPSInit,sizeof(VDPStatus));
  memcpy(Palette,PalInit,sizeof(Palette));
  memcpy(PSG,PSGInit,sizeof(PSG));

  VKey=PKey=1;
  ChrTab=ColTab=ChrGen=VPAGE=VRAM;
  FGColor=BGColor=ScrMode=0;
  R.PC.W=0;R.SP.W=0xF000;R.IFF=0;

  if(Verbose) printf("OK\nRUNNING ROM CODE...\n");
  A=Z80(RAM,R);

  if(Verbose) printf("EXITED at PC = %Xh.\n",A);
  return(1);
}



/****************************************************************/
/*** Free memory allocated with StartMSX().                   ***/
/****************************************************************/
void TrashMSX()
{
  byte I,J,K;
  FILE *F;
  
  if(RAM)  free(RAM);
  if(VRAM) free(VRAM);

  for(I=0;I<4;I++)
    for(J=0;J<((I==3)? 4:1);J++)
      for(K=0;K<4;K++)
        if(MemMap[I][J][K]) free(MemMap[I][J][K]);

  if(SaveCMOS)
  {
    if(Verbose) printf("Writing CMOS.ROM...");
    if(F=fopen("CMOS.ROM","wb"))
      if(fwrite(RTC,1,sizeof(RTC),F)!=sizeof(RTC))
      { fclose(F);F=0; }
    if(F) { fclose(F);PRINTOK; }
    else PRINTFAILED;
  }

  if(PrnStream) fclose(PrnStream);
}



/****************************************************************/
/*** Load a cartridge ROM from .ROM file into given slot.     ***/
/****************************************************************/
int LoadCart(Name,Slot)
  char *Name;
  int Slot;
{
  static char Text[] = "  Cartridge # ROM #";
  FILE *F;
  int C1,C2;

  if(!(F=fopen(Name,"rb"))) return(0);
  if(Verbose) printf("Found %s:\n",Name);

  C1=fgetc(F);C2=fgetc(F);
  if((C1!='A')||(C2!='B'))
  { 
    if(Verbose) printf("  Not a valid cartridge ROM\n");
    fclose(F);return(0);
  }

  C2=fgetc(F);C2=(fgetc(F)>>6)&0x03;
  for(rewind(F);C2<3;C2++)
  {
    C1=fgetc(F);
    if(C1<0) break; else ungetc(C1,F);
    Text[12]='0'+Slot;Text[18]='0'+C2;
    LoadROM(Text,F,Slot,0,C2);
  }
  C1=fgetc(F);

  if((C2==3)&&(C1>=0))
  {
    if(Verbose) printf("  Loading MegaROM blocks...");
    ungetc(C1,F);C1=Slot-1;
    for(C2=0;C2<4;C2++)
      if(!(ROMMap[C1][C2]=malloc(PAGESIZE/2))) break;
    if(C2<4)
    { fclose(F);PRINTFAILED;return(0); }

    memcpy(ROMMap[C1][0],MemMap[Slot][0][1],PAGESIZE/2);
    memcpy(ROMMap[C1][2],MemMap[Slot][0][2],PAGESIZE/2);
    memcpy(ROMMap[C1][1],MemMap[Slot][0][1]+PAGESIZE/2,PAGESIZE/2);
    memcpy(ROMMap[C1][3],MemMap[Slot][0][2]+PAGESIZE/2,PAGESIZE/2);

    for(Slot--;C2<256;C2++)
    {
      C1=fgetc(F);
      if(C1<0) break; else ungetc(C1,F);
      if(!(ROMMap[Slot][C2]=malloc(PAGESIZE/2))) break;
      if(fread(ROMMap[Slot][C2],1,PAGESIZE/2,F)!=PAGESIZE/2)
      { free(ROMMap[Slot][C2]);ROMMap[Slot][C2]=NULL;break; }
    }

    if(Verbose) printf("%d blocks loaded\n",C2);    
  }
  
  fclose(F);return(1);
}



/****************************************************************/
/*** Load a ROM page from the file F into slot P, subslot S,  ***/
/*** page A. Return memory address of a page or 0 if failed.  ***/
/****************************************************************/
byte *LoadROM(STR,F,P,S,A)
  char *STR;
  FILE *F;
  int P,S,A;
{
  if(Verbose) printf("%s: allocating memory...",STR);
  if(MemMap[P][S][A]) free(MemMap[P][S][A]);
  if(MemMap[P][S][A]=malloc(PAGESIZE))
  {
    if(Verbose) printf("loading...");
    if(fread(MemMap[P][S][A],1,PAGESIZE,F)!=PAGESIZE)
    { free(MemMap[P][S][A]);MemMap[P][S][A]=0; }
  }
  if(P!=3) 
    MemMap[P][0][A]=MemMap[P][1][A]=
    MemMap[P][2][A]=MemMap[P][3][A]=MemMap[P][S][A];
  if(Verbose) printf(MemMap[P][S][A]? MsgOK:MsgFAILED);  
  return(MemMap[P][S][A]);
}



/****************************************************************/
/*** Write number into given VDP register.                    ***/
/****************************************************************/
void VDPOut(register byte R,register byte V)
{ 
  register byte J,*P;  

  switch(R)  
  {
    case  0: J=ScrMode;
             switch(((V&0x0E)>>1)|(VDP[1]&0x18))
             {
               case 0x10: J=0;break;
               case 0x00: J=1;break;
               case 0x01: J=2;break;
               case 0x08: J=3;break;
               case 0x02: J=4;break;
               case 0x03: J=5;break;
               case 0x04: J=6;break;
               case 0x05: J=7;break;
               case 0x07: J=8;break;
               case 0x12: J=MAXSCREEN+1;break;
             }   
             if(J!=ScrMode)
             {
               ChrTab=
                 VRAM+((long)(VDP[2]&SCR[ScrMode].R2)<<((ScrMode>6)? 11:10));
               ChrGen=VRAM+((long)(VDP[4]&SCR[ScrMode].R4)<<11);
               ColTab=VRAM+((long)(VDP[3]&SCR[ScrMode].R3)<<6)
                          +((long)VDP[10]<<14);
	       SprTab=VRAM+((long)(VDP[5]&SCR[ScrMode].R5)<<7)
                          +((long)VDP[11]<<15);
               SprGen=VRAM+((long)VDP[6]<<11);
               SCR[ScrMode=J].Init();SCR[J].Colors();
             }
             break;
    case  1: J=ScrMode;
             switch(((VDP[0]&0x0E)>>1)|(V&0x18))
             {
               case 0x10: J=0;break;
               case 0x00: J=1;break;
               case 0x01: J=2;break;
               case 0x08: J=3;break;
               case 0x02: J=4;break;
               case 0x03: J=5;break;
               case 0x04: J=6;break;
               case 0x05: J=7;break;
               case 0x07: J=8;break;
               case 0x12: J=MAXSCREEN+1;break;
             }   
             if(J!=ScrMode)
             {
               ChrTab=
                 VRAM+((long)(VDP[2]&SCR[ScrMode].R2)<<((ScrMode>6)? 11:10));
               ChrGen=VRAM+((long)(VDP[4]&SCR[ScrMode].R4)<<11);
               ColTab=VRAM+((long)(VDP[3]&SCR[ScrMode].R3)<<6)
                          +((long)VDP[10]<<14);
               SprTab=VRAM+((long)(VDP[5]&SCR[ScrMode].R5)<<7)  
                          +((long)VDP[11]<<15);
               SprGen=VRAM+((long)VDP[6]<<11);
               SCR[ScrMode=J].Init();SCR[J].Colors();
             }
             break;       
    case  2: ChrTab=
               VRAM+((long)(V&SCR[ScrMode].R2)<<((ScrMode>6)? 11:10));
             break;
    case  3: ColTab=
               VRAM+((long)(V&SCR[ScrMode].R3)<<6)+((long)VDP[10]<<14);
             break;
    case  4: ChrGen=VRAM+((long)(V&SCR[ScrMode].R4)<<11);
             break;
    case  5: SprTab=
               VRAM+((long)(V&SCR[ScrMode].R5)<<7)+((long)VDP[11]<<15);
             break;
    case  6: V&=0x3F;SprGen=VRAM+((long)V<<11);
             break;
    case  7: FGColor=V>>4;BGColor=V&0x0F;
             if(!VDP[13]) { XBGColor=BGColor;XFGColor=FGColor; }
             SCR[ScrMode].Colors();
             break;
    case  8:
    case  9: VDP[R]=V;SCR[ScrMode].Colors();break;
    case 10: V&=0x07;
             ColTab=
               VRAM+((long)(VDP[3]&SCR[ScrMode].R3)<<6)+((long)V<<14);
             break;
    case 11: V&=0x03;
             SprTab=
               VRAM+((long)(VDP[5]&SCR[ScrMode].R5)<<7)+((long)V<<15);
             break;
    case 13: if(!V) { XBGColor=BGColor;XFGColor=FGColor; }
             break;
    case 14: V&=0x07;VPAGE=(V<VRAMPages)? VRAM+((long)V<<14):NULL;
             break;
    case 15: V&=0x0F;break;
    case 16: V&=0x0F;break;
    case 17: V&=0xBF;break;
    case 33:
    case 37:
    case 41: V&=0x01;break;
    case 35:
    case 39:
    case 43: V&=0x03;break;
    case 44: if(VDPStatus[2]&0x80)
             {
               if(MMC.Mode)
               {
                 if(V||MMC.TR)
                 {
                   P=MMC.P+MMC.XC/MMC.Diff;
                   J=MMC.Shift[MMC.XC%MMC.Diff];
                   *P=(*P&~(MMC.Mask<<J))|Oper(V<<J,*P,MMC.Op);
                 }
                 MMC.NXC--;
               }
               else
               {
                 P=MMC.P+MMC.XC/MMC.Diff;
                 *P=V;
                 MMC.NXC-=2;
               }
               if(MMC.NXC>0) MMC.XC+=MMC.XStep;
               else
                 if(--MMC.NY)
                 { MMC.XC=MMC.X;MMC.NXC=MMC.NX;MMC.P+=MMC.YStep; }
                 else VDPStatus[2]&=0x7E;
             }
             break;
    case 46: if((ScrMode>4)&&(ScrMode<9)) DoVDP(V);
             break;
  }
  VDP[R]=V;return;
} 



/****************************************************************/
/*** This function is used by DoVDP() for operations on       ***/
/*** pixels.                                                  ***/
/****************************************************************/
byte Oper(register byte A,register byte B,register byte N)
{
  switch(N)
  {
    case 0:  return(A);
    case 1:  return(A&B);
    case 2:  return(A|B);
    case 3:  return(A^B);
    case 4:  return(~A);
  }
  return(B);
}



/****************************************************************/
/*** Perform given V9938 operation.                           ***/
/****************************************************************/
void DoVDP(register byte V)
{
  int SX,SY,DX,DY,NX,NY,AX,AY,TX,TY,L;
  register byte *T,*P,SM,I,J,TR;

  byte Shift[4][4] = { {4,0,0,0},{6,4,2,0},{4,0,0,0},{0,0,0,0} };
  byte Diff[4]     = { 2,4,2,1 };
  byte Mask[4]     = { 0x0F,0x03,0x0F,0xFF };
  int  Width[4]    = { 256,512,512,256 };
  byte Line[4]     = { 7,7,8,8 };

  SX=VDP[32]+((word)VDP[33]<<8);SY=VDP[34]+((word)VDP[35]<<8);
  DX=VDP[36]+((word)VDP[37]<<8);DY=VDP[38]+((word)VDP[39]<<8);
  NX=VDP[40]+((word)VDP[41]<<8);NY=VDP[42]+((word)VDP[43]<<8);
  SM=ScrMode-5;TR=!(V&0x08);

  switch(V&0xF0)
  {
    /*** Get color of a point at SX,SY: ***/
    case 0x40: P=VRAM+((long)SY<<Line[SM])+SX/Diff[SM];
               VDPStatus[7]=(*P>>Shift[SM][SX%Diff[SM]])&Mask[SM];
               break;
    /*** Set color of a point at DX,DY: ***/
    case 0x50: J=VDP[44]&Mask[SM];V&=0x0F;if(!J&&(V>7)) break;
               P=VRAM+((long)DY<<Line[SM])+DX/Diff[SM];
               I=Shift[SM][DX%Diff[SM]];
               *P=(*P&~(Mask[SM]<<I))|Oper(J<<I,*P,V);
               break;
    /*** Search in given dir. for a point of given color from SX,SY: ***/
    case 0x60: P=VRAM+((long)SY<<Line[SM]);
               VDP[44]&=Mask[SM];I=(VDP[45]&0x02)>>1;
               if(VDP[45]&0x04)
                 for(;SX>=0;SX--)
                 {                   

                   T=P+SX/Diff[SM];
                   J=(*T>>Shift[SM][SX%Diff[SM]])&Mask[SM];
                   if(J==VDP[44]) break;
                   if((J!=BGColor)^I) break;
                 }
               else
                 for(;SX<Width[SM];SX++)
                 {
                   T=P+SX/Diff[SM];
                   J=(*T>>Shift[SM][SX%Diff[SM]])&Mask[SM];
                   if(J==VDP[44]) break;
                   if((J!=BGColor)^I) break;
                 }
               SX=(SX>=Width[SM])? Width[SM]-1:(SX<0)? 0:SX;
               VDPStatus[8]=SX&0xFF;VDPStatus[9]=(SX>>8)|0xFE;
               if(J==VDP[44]) VDPStatus[2]&=0xEF;
               else           VDPStatus[2]|=0x10;
               break;
    /*** Draw a line of length NX,NY from DX,DY: ***/
    case 0x70: J=VDP[44]&Mask[SM];V&=0x0F;if(!J&&(V>7)) break;
               if(VDP[45]&0x01) { L=NX;NX=NY;NY=L; }
               TX=VDP[45]&0x04? -1:1; TX=NX? TX:0; SX=DX+(NX+1)*TX; AX=2*NX;
               TY=VDP[45]&0x08? -1:1; TY=NY? TY:0; SY=DY+(NY+1)*TY; AY=2*NY;
               if(VDP[45]&0x01)
                 for(L=AX-NY;DY!=SY;DY+=TY,L+=AX)
                 {
                   P=VRAM+((long)DY<<Line[SM])+DX/Diff[SM];
                   I=Shift[SM][DX%Diff[SM]];
                   *P=(*P&~(Mask[SM]<<I))|Oper(J<<I,*P,V);
                   if(L>=0) { L-=AY;DX+=TX; }
                 }
               else
                 for(L=AY-NX;DX!=SX;DX+=TX,L+=AY)
                 {
                   P=VRAM+((long)DY<<Line[SM])+DX/Diff[SM];
                   I=Shift[SM][DX%Diff[SM]];
                   *P=(*P&~(Mask[SM]<<I))|Oper(J<<I,*P,V);
                   if(L>=0) { L-=AX;DY+=TY; }
                 }
               break;                
    /*** Fill a box with a given color: ***/
    case 0xC0: V&=0xF0;
    case 0x80: J=VDP[44]&Mask[SM];V&=0x0F;if(!J&&(V>7)) break;
               if(VDP[45]&0x04) SX=DX-NX; else { SX=DX;DX+=NX; }
               if(VDP[45]&0x08) SY=DY-NY; else { SY=DY;DY+=NY; }
               for(;SY<DY;SY++)
                 for(L=SX;L<DX;L++)
                 {
                   P=VRAM+((long)SY<<Line[SM])+L/Diff[SM];
                   I=Shift[SM][L%Diff[SM]];
                   *P=(*P&~(Mask[SM]<<I))|Oper(J<<I,*P,V);
                 }         
               break;
    /*** Copy a box from one place to another: ***/
    case 0xE0: NX=VDP[45]&0x08? DX:Width[SM]-DX;
               TX=VDP[45]&0x04? -1:1;TY=VDP[45]&0x08? -1:1;
               for(;NY;NY--,SY+=TY,DY+=TY)
                 for(L=NX,AX=SX;L;L--,AX+=TX)
                 {
                   T=VRAM+AX/Diff[SM];I=Shift[SM][AX%Diff[SM]];
                   P=T+((long)SY<<Line[SM]);
                   J=(*P>>I)&Mask[SM];
                   P=T+((long)DY<<Line[SM]);
                   *P=(*P&~(Mask[SM]<<I))|(J<<I);
                 }   
               break;
    case 0xD0: TX=VDP[45]&0x04? -1:1;TY=VDP[45]&0x08? -1:1;
/*printf("%d,%d -> %d,%d [%dx%d]\n",SX,SY,DX,DY,NX,NY);*/
               for(;NY;NY--,SY+=TY,DY+=TY)
                 for(L=NX,AX=SX,AY=DX;L;L--,AX+=TX,AY+=TX)
                 {
                   P=VRAM+((long)SY<<Line[SM])+AX/Diff[SM];
                   J=(*P>>Shift[SM][AX%Diff[SM]])&Mask[SM];
                   P=VRAM+((long)DY<<Line[SM])+AY/Diff[SM];
                   I=Shift[SM][AY%Diff[SM]];
                   *P=(*P&~(Mask[SM]<<I))|(J<<I);
                 }   
               break;
    case 0x90: V&=0x07;TX=VDP[45]&0x04? -1:1;TY=VDP[45]&0x08? -1:1;
               for(;NY;NY--,SY+=TY,DY+=TY)
                 for(L=NX,AX=SX,AY=DX;L;L--,AX+=TX,AY+=TX)
                 {
                   P=VRAM+((long)SY<<Line[SM])+AX/Diff[SM];
                   J=(*P>>Shift[SM][AX%Diff[SM]])&Mask[SM];
                   if(J||TR)
                   {                   
                     P=VRAM+((long)DY<<Line[SM])+AY/Diff[SM];
                     I=Shift[SM][AY%Diff[SM]];
                     *P=(*P&~(Mask[SM]<<I))|Oper(J<<I,*P,V);
                   }
                 }
               break;
    /*** Transfer a box of data from CPU: ***/
    case 0xF0: V&=0xF0;
    case 0xB0: VDPStatus[2]|=0x81;
               MMC.Mode=((V&0xF0)==0xB0);
               MMC.TR=!(V&0x08);MMC.Op=V&0x07;
               MMC.P=VRAM+((long)DY<<Line[SM]);
               MMC.YStep=(VDP[45]&0x08? -1:1)<<Line[SM];
               MMC.XStep=(MMC.Mode? 1:2)*(VDP[45]&0x04? -1:1);
               MMC.X=MMC.XC=DX;MMC.NX=MMC.NXC=NX;MMC.NY=NY;
               MMC.Diff=Diff[SM];MMC.Mask=Mask[SM];
               memcpy(MMC.Shift,Shift[SM],4);
               VDPOut(44,VDP[44]);
               break;
    default: printf("Unimplemented V9938 operation %X\n",V);
  }
}



/****************************************************************/
/*** Send a character to the printer.                         ***/
/****************************************************************/
void Printer(byte V) { fputc(V,PrnStream); }



/****************************************************************/
/*** Write number into given PSG register.                    ***/
/****************************************************************/
void PSGOut(register byte R,register byte V)
{
  switch(R)
  {
    case  6:
    case  8:
    case  9:
    case 10: V&=0x1F;break;
    case 13: V&=0x0F;break;
  }
  PSG[R]=V;
}



/****************************************************************/
/*** Read number from given RTC register.                     ***/
/****************************************************************/
byte RTCIn(register byte R)
{
  register byte J;
  time_t T;
  char *S;

  /*** ctime() returns "DDD MMM dd hh:mm:ss YYYYn0" ***/
  static byte RTable[13] =
  { 18,17,15,14,12,11,25,9,8,25,25,23,22 };
  static char *Days[7] =
  { "Sun","Mon","Tue","Wed","Thu","Fri","Sat" }; 
  static char *Months[12] =
  {
    "Jan","Feb","Mar","Apr","May","Jun",
    "Jul","Aug","Sep","Oct","Nov","Dec"
  };

  J=RTCMode&0x03;
  if(R>12) J=(R==13)? RTCMode:NORAM;
  else
    if(J) J=RTC[J][R];
    else
    {
      time(&T);S=ctime(&T);
      switch(R)
      {
        case 6:  for(J=0;(J<7)&&strncmp(S,Days[J],3);J++);
                 J=(J==7)? 0:J+1;break;
        case 9:  for(J=0;(J<12)&&strncmp(S+4,Months[J],3);J++);
                 J=(J==12)? 0:(J+1)%10;break;
        case 10: for(J=0;(J<12)&&strncmp(S+4,Months[J],3);J++);
                 J=(J==12)? 0:(J+1)/10;break;
        case 11: S[24]='\0';J=(atoi(S+20)-1980)%10;break;
        case 12: S[24]='\0';J=((atoi(S+20)-1980)/10)%10;break;
        default: J=S[RTable[R]]-'0';
      }
    }
  return(J|0xF0);
}



/****************************************************************/
/*** Set sound parameters, refresh screen, check keyboard and ***/
/*** sprites. Call this function on each interrupt.           ***/
/****************************************************************/
void Interrupt(void)
{
  static byte BFlag=0,BCount=0;
  static int UCount=1;
  register byte I,J;

  /*** Setting sound parameters: ***/
  for(J=I=0;I<3;I++,J+=2)
    if((PSG[J]|PSG[J+1])&&(PSG[7]&(0x01<<I)))
    {
      Freq[I]=SNDCONST/32/(((word)PSG[J+1]<<8)+PSG[J]);
      Volume[I]=PSG[I+8]&0x10? 128:PSG[I+8]<<4;
    }
    else Volume[I]=0;

  /*** Calling sound routine: ***/
  Sound();

  /*** Checking keyboard: ***/
  Keyboard();

  if(!--UCount)
  {
    UCount=UPeriod;
  
    /*** Blinking for TEXT80: ***/
    if(BCount) BCount--;
    else
    {
      BFlag=!BFlag;
      BCount=BFlag? VDP[13]>>4:VDP[13]&0x0F;
      if(BCount)
        if(BFlag)
        { XFGColor=VDP[12]>>4;XBGColor=VDP[12]&0x0F; }
        else
        { XFGColor=FGColor;XBGColor=BGColor; }
    }

    /*** Refreshing screen: ***/
    (SCR[ScrMode].Refresh)();

    /*** Checking sprites: ***/
    if(!SpritesOFF&&ScrMode&&(ScrMode<MAXSCREEN+1)) CheckSprites();
  }
}



/****************************************************************/
/*** Check for sprite collisions and 5th/9th sprite in a row. ***/
/****************************************************************/
void CheckSprites(void)
{
  register char DH,DV,SH,SV;
  register byte I,J,N,M,*PS,*PD,*S,*D;

  VDPStatus[0]&=0xB0;
  M=(ScrMode>3)? 8:4;

  if(Sprites16x16)
    for(J=0,S=SprTab;J<32;J++,S+=4)
    {
      N=0;
      for(I=J+1,D=S+4;I<32;I++,D+=4)
      {
        DV=S[0]-D[0];DV*=(SV=(DV<0)? -1:1);
        if(DV<16)
	{
          N++;DH=S[1]-D[1];DH*=(SH=(DH<0)? -1:1);
          if(DH<16)
	  {
            PS=SprGen+((long)(S[2]&0xFC)<<3);
            PD=SprGen+((long)(D[2]&0xFC)<<3);
            if(SV>0) PD+=DV; else PS+=DV;
            if(SH>0)
              while
              (
                (DV<16)&&!((long)*PS&((long)*PD<<DH))&&
                !((long)*(PS+16)&((long)*(PD+16)<<DH))
              ) { DV++;PS++;PD++; }
            else
              while
              (
                (DV<16)&&!((long)*PD&((long)*PS<<DH))&&
                !((long)*(PD+16)&((long)*(PS+16)<<DH))
              ) { DV++;PS++;PD++; }
            if(DV<16) VDPStatus[0]|=0x20;     
          }
        }
      }  
      if(M>4) VDPStatus[0]=(VDPStatus[0]&0xF0)|0x40|(J&0x0F);
    }
  else
    for(J=0,S=SprTab;J<32;J++,S+=4)
    {
      N=0;
      for(I=J+1,D=S+4;I<32;I++,D+=4)
      {
        DV=S[0]-D[0];DV*=(SV=(DV<0)? -1:1);
        if(DV<8)
        {
          N++;DH=S[1]-D[1];DH*=(SH=(DH<0)? -1:1);
          if(DH<8)
          {
            PS=SprGen+((long)S[2]<<3);PD=SprGen+((long)D[2]<<3);
            if(SV>0) PD+=DV; else PS+=DV;
            if(SH>0)
              while((DV<8)&&!((long)*PS&((long)*PD<<DH)))
              { DV++;PS++;PD++; }
            else
              while((DV<8)&&!((long)*PD&((long)*PS<<DH)))
              { DV++;PS++;PD++; }
            if(DV<8) VDPStatus[0]|=0x20;
          }
        }
      }
      if(M>4) VDPStatus[0]=(VDPStatus[0]&0xF0)|0x40|(J&0x0F);
    }
}
